#include "DiagInterface.h"

#define DEV_VID 0x0481
#define DEV_PID 0x0002
#define DIAG_IFACE_NUM 1
#define DIAG_IFACE_IN_EP 0x83
#define DIAG_IFACE_TIMEOUT 3000
#define KBD_COLS 23
#define KBD_ROWS 4

using namespace std;

/*
 *
 */
DiagInterface::DiagInterface(void): dev(NULL)
{
    int err;

    if ((err = libusb_init(NULL)) != 0)
        throw runtime_error(string("error initialising libusb: ") +
                libusb_error_name(err));
    libusb_set_debug(NULL, 3);

    dev = openDevice();

    setState(drNothing, false);
}

/*
 *
 */
DiagInterface::~DiagInterface(void)
{
    closeDevice();
    libusb_exit(NULL);
}

/*
 *
 */
unsigned short DiagInterface::vref(void)
{
    int val = -1;

    /* wait for reported vref to settle---not sure why first read sometimes
     * appears as old result, despite it changing on controller---no double
     * some misunderstanding about how LUFA or libusb works on my part
     */
    for (int i = 0; i < 16; i++)
    {
        int newVal;

        setState(drNothing, false);
        setState(drVref);
        newVal = currState[1] + (currState[2] << 8);

        if (val == newVal)
            break;
        val = newVal;

        if (i == 15)
            throw runtime_error("error: too many retries waiting for vref to "
                    "settle");
    }

    return val;
}

/*
 *
 */
void DiagInterface::setVref(unsigned short val)
{
    sendCmd(dcSetVref, val);
}

/*
 *
 */
void DiagInterface::storeVref(void)
{
    sendCmd(dcStoreVref, 1);
}

/*
 *
 */
void DiagInterface::autoCalibrate(void)
{
    sendCmd(dcAutoCalibrate, 1);
}

/*
 *
 */
static void populateCol(vector<vector<bool>> &states, size_t col,
        unsigned char val)
{
    if (col >= KBD_COLS)
        return;

    for (int i = 0; i < KBD_ROWS; i++)
        states[col][i] = val & (1 << i);
}
std::vector<std::vector<bool>> DiagInterface::keyStates(void)
{
    vector<vector<bool>> states(KBD_COLS, vector<bool>(KBD_ROWS, false));
    bool got1stCols = false;
    bool got2ndCols = false;

    setState(drNothing, false);
    setState(dr1stCols);
    while (!got1stCols || !got2ndCols)
    {
        if (currState[0] == dr1stCols)
        {
            got1stCols = true;
            for (int i = 1, col = 0; i < 7; i++)
            {
                populateCol(states, col++, currState[i] & 0xf);
                populateCol(states, col++, currState[i] >> 4);
            }
        }
        else if (currState[0] == dr2ndCols)
        {
            got2ndCols = true;
            for (int i = 1, col = 12; i < 7; i++)
            {
                populateCol(states, col++, currState[i] & 0xf);
                populateCol(states, col++, currState[i] >> 4);
            }
        }

        updateState();
    }

    return states;
}

/*
 *
 */
void DiagInterface::enterBootloader(void)
{
    sendCmd(dcJumpBootloader, 1);
}

/*
 *
 */
std::vector<std::vector<unsigned char>> DiagInterface::scancodes(void)
{
    vector<vector<unsigned char>> sc(KBD_COLS,
            vector<unsigned char>(KBD_ROWS, 0x00));

    setState(drNothing, false);
    setState(drScanCodes);
    for (int i = 0; i < KBD_COLS; i++)
    {
        for (int i = 0; i < 4; i++)
            sc[currState[1]][i] = currState[i + 2];
        updateState();
    }

    return sc;
}

/*
 *
 */
void DiagInterface::setScancodes(std::vector<std::vector<unsigned char>> codes)
{
    if (codes.size() != KBD_COLS)
        throw runtime_error("error setting scancodes: number of columns "
                "differs from keyboard");
    if (codes[0].size() != KBD_ROWS)
        throw runtime_error("error setting scancodes: number of rows "
                "differs from keyboard");

    setState(drNothing, false);
    for (int i = 0; i < KBD_COLS; i++)
        for (int j = 0; j < KBD_ROWS; j++)
            setScancode(i, j, codes[i][j]);
}

/*
 *
 */
void DiagInterface::setScancode(int col, int row, unsigned char scancode)
{
    if (col < 0 || col >= KBD_COLS)
        throw runtime_error("error setting scancode: column out of range "
                "(min=0, max=" + to_string(KBD_COLS) + ", val=" +
                to_string(col) + ")");
    if (row < 0 || row >= KBD_ROWS)
        throw runtime_error("error setting scancode: row out of range "
                "(min=0, max=" + to_string(KBD_ROWS) + ", val=" +
                to_string(row) + ")");
    sendCmd(dcSetScanCode, col, row, scancode);
}

/*
 *
 */
void DiagInterface::loadScancodes(void)
{
    sendCmd(dcLoadScanCodes, 1);
}

/*
 *
 */
void DiagInterface::storeScancodes(void)
{
    sendCmd(dcStoreScanCodes, 1);
}

/*
 *
 */
libusb_device_handle * DiagInterface::openDevice(void)
{
    int err;

    libusb_device_handle *dev =
        libusb_open_device_with_vid_pid(NULL, DEV_VID, DEV_PID);
    if (dev == NULL)
        throw runtime_error("error: could not open device");

    err = libusb_kernel_driver_active(dev, DIAG_IFACE_NUM);
    if (err == 1)
    {
        if ((err = libusb_detach_kernel_driver(dev, DIAG_IFACE_NUM)) != 0)
            throw runtime_error(string("error: could not detach kernel driver "
                        "from interface: ") + libusb_error_name(err));
    }
    else if (err != 0)
        throw runtime_error(string("error: could not check if kernel driver "
                    "active: ") + libusb_error_name(err));

    if ((err = libusb_claim_interface(dev, DIAG_IFACE_NUM)) != 0)
        throw runtime_error(string("error: could not claim diag interface: ") +
                libusb_error_name(err));

    return dev;
}

/*
 *
 */
void DiagInterface::closeDevice(void)
{
    int err;

    if ((err = libusb_release_interface(dev, DIAG_IFACE_NUM)) != 0)
        cerr << "warning: had error whilst releasing diag interface: " <<
            libusb_error_name(err) << endl;
    libusb_close(dev);
}

/*
 *
 */
void DiagInterface::sendCmd(DiagReportCommand cmd, unsigned char p1,
        unsigned char p2, unsigned char p3)
{
    unsigned char buf[8];
    buf[0] = cmd;
    buf[1] = p1;
    buf[2] = p2;
    buf[3] = p3;
    sendCtrl(buf);
}

/*
 *
 */
void DiagInterface::sendCmd(DiagReportCommand cmd, unsigned int data)
{
    unsigned char buf[8];
    buf[0] = cmd;
    buf[1] = data & 0xff;
    buf[2] = data >> 8;
    buf[3] = data >> 16;
    buf[4] = data >> 24;
    sendCtrl(buf);
}

/*
 *
 */
void DiagInterface::sendCtrl(unsigned char buf[8])
{
    int err;

    if ((err = libusb_control_transfer(dev, LIBUSB_REQUEST_TYPE_CLASS |
                    LIBUSB_RECIPIENT_INTERFACE | LIBUSB_ENDPOINT_OUT, 0x09,
                    (2 << 8) | 0, 1, buf, sizeof(buf), DIAG_IFACE_TIMEOUT)) !=
            sizeof(buf))
        throw runtime_error(string("error during control transfer: ") +
                libusb_error_name(err));
}

/*
 *
 */
void DiagInterface::setState(DiagReportState state, bool block)
{
    sendCmd(dcSetState, state);

    if (block)
    {
        do
        {
            updateState();
        } while (currState[0] != state);
    }
}

/*
 *
 */
void DiagInterface::updateState(void)
{
    int err, bytesTxd;
    unsigned char buf[8];

    memset(currState, 0x00, sizeof(currState));

    if ((err = libusb_interrupt_transfer(dev, DIAG_IFACE_IN_EP, buf,
                    sizeof(buf), &bytesTxd, DIAG_IFACE_TIMEOUT)) != 0)
        throw runtime_error(string("error during interrupt transfer: ") +
                libusb_error_name(err));

    memcpy(currState, buf, bytesTxd);
}
