#include "Frontend.h"

using namespace std;

/*
 *
 */
Frontend::Frontend(DiagInterface &diag, QWidget *parent):
    QWidget(parent),
    diag(diag)
{
    setWindowTitle("Beamspring USB Util");

    /*
     * vref stuff
     */
    vrefMaskTimer = new QTimer(this);
    vrefMaskTimer->setSingleShot(true);

    QGroupBox *vrefGroup = new QGroupBox("Voltage threshold");

    QLabel *vrefLabel = new QLabel("<p>The threshold that a key is considered "
            "pressed or not.</p>"
            "<p>If a valid set of scancodes is loaded (some keys "
            "must be set to PRESSED or RELEASED) then in most cases the "
            "threshold can be auto-calibrated when the keyboard "
            "is plugged in (this happens if 65535 is stored in the "
            "controller's EEPROM).</p>"
            "<p>Alternatively, a forced override value can "
            "be stored in the controller's EEPROM.</p>");
    vrefLabel->setWordWrap(true);

    QLabel *vrefSpinBoxLabel = new QLabel("Current threshold:");
    vrefSpinBox = new QSpinBox;
    vrefSpinBox->setMaximum(65535);
    QHBoxLayout *vrefSpinBoxHBox = new QHBoxLayout;
    vrefSpinBoxHBox->addWidget(vrefSpinBoxLabel);
    vrefSpinBoxHBox->addWidget(vrefSpinBox);

    autoCalButton = new QPushButton;
    storeVrefButton =
        new QPushButton("Store current threshold as override in EEPROM");

    QVBoxLayout *vrefVBox = new QVBoxLayout;
    vrefVBox->addWidget(vrefLabel);
    vrefVBox->addLayout(vrefSpinBoxHBox);
    vrefVBox->addWidget(autoCalButton);
    vrefVBox->addWidget(storeVrefButton);

    vrefGroup->setLayout(vrefVBox);

    /*
     * matrix stuff
     */
    QTimer *keyStatesTimer = new QTimer(this);
    keyStatesTimer->start(0);

    QGroupBox *matrixGroup = new QGroupBox("Layout");

    matrixGrid = new QGridLayout;
    matrixGrid->setSpacing(0);

    matrixWidget = new QWidget;
    matrixWidget->setLayout(matrixGrid);

    matrixScrollArea = new QScrollArea;
    matrixScrollArea->setWidget(matrixWidget);
    matrixScrollArea->setWidgetResizable(true);
    matrixScrollArea->setMinimumHeight(200);
    matrixScrollArea->setMinimumWidth(800);

    QScroller::grabGesture(matrixScrollArea,
            QScroller::LeftMouseButtonGesture); // heh cool

    loadMatrixButton = new QPushButton("Load layout from EEPROM");
    storeMatrixButton = new QPushButton("Store layout in EEPROM");

    QPushButton *importMatrixButton = new QPushButton("Import layout");
    QPushButton *exportMatrixButton = new QPushButton("Export layout");

    QHBoxLayout *layoutHBox = new QHBoxLayout;
    layoutHBox->addWidget(loadMatrixButton);
    layoutHBox->addWidget(storeMatrixButton);
    layoutHBox->addWidget(importMatrixButton);
    layoutHBox->addWidget(exportMatrixButton);

    QVBoxLayout *matrixLayout = new QVBoxLayout;
    matrixLayout->addLayout(layoutHBox);
    matrixLayout->addWidget(matrixScrollArea);

    matrixGroup->setLayout(matrixLayout);

    /*
     * main layout
     */
    QHBoxLayout *hbox1 = new QHBoxLayout;
    hbox1->addWidget(vrefGroup, 2);
    hbox1->addStretch(1);

    QVBoxLayout *vbox1 = new QVBoxLayout;
    vbox1->addLayout(hbox1);
    vbox1->addWidget(matrixGroup);

    setLayout(vbox1);

    /* ask for a couple of things immediately */
    QTimer::singleShot(0, this, SLOT(updateVref()));
    QTimer::singleShot(0, this, SLOT(buildMatrix()));

    connect(vrefSpinBox, SIGNAL(valueChanged(int)),
            SLOT(vrefValueChanged(int)));
    connect(vrefMaskTimer, SIGNAL(timeout()), SLOT(setVrefFromBox()));
    connect(autoCalButton, SIGNAL(clicked()), SLOT(autoCalButtonClicked()));
    connect(storeVrefButton, SIGNAL(clicked()), SLOT(storeVrefButtonClicked()));
    connect(keyStatesTimer, SIGNAL(timeout()), SLOT(updateKeyStates()));
    connect(storeMatrixButton, SIGNAL(clicked()),
            SLOT(storeMatrixButtonClicked()));
    connect(loadMatrixButton, SIGNAL(clicked()),
            SLOT(loadMatrixButtonClicked()));
    connect(importMatrixButton, SIGNAL(clicked()),
            SLOT(importMatrixButtonClicked()));
    connect(exportMatrixButton, SIGNAL(clicked()),
            SLOT(exportMatrixButtonClicked()));
}

/*
 *
 */
Frontend::~Frontend(void)
{
}

/*
 *
 */
void Frontend::updateVref(void)
{
    disconnect(vrefSpinBox, SIGNAL(valueChanged(int)), this,
            SLOT(vrefValueChanged(int)));

    vrefSpinBox->setValue(diag.vref());

    connect(vrefSpinBox, SIGNAL(valueChanged(int)), this,
            SLOT(vrefValueChanged(int)));

    /* re-enable widgets and set autocal button's text---should probably use
     * state machine or something as this is not needed every time
     */
    setEnabled(true);
    autoCalButton->setText("Auto-calibrate");
}

/*
 *
 */
void Frontend::vrefValueChanged(int)
{
    /* mask for a bit before actually setting value in case just mashing one
     * of the arrows, or typing etc.
     */
    vrefMaskTimer->start(500);
    autoCalButton->setEnabled(false);
    storeVrefButton->setEnabled(false);
}

/*
 *
 */
void Frontend::setVrefFromBox(void)
{
    qDebug() << "setting vref";
    diag.setVref(vrefSpinBox->value());
    autoCalButton->setEnabled(true);
    storeVrefButton->setEnabled(true);
}

/*
 *
 */
void Frontend::autoCalButtonClicked(void)
{
    setEnabled(false);
    autoCalButton->setText("Auto-calibrating...");

    diag.autoCalibrate();

    /* keyboard will disappear for a bit while it does its thing */
    QTimer::singleShot(1000, this, SLOT(autoCalComplete()));
}

/*
 *
 */
void Frontend::autoCalComplete(void)
{
    unsigned short vref = diag.vref();

    QMessageBox::StandardButton result = QMessageBox::question(this,
            "Auto-calibration complete",
            "Auto-calibration finished, threshold is " +
            QString::number(vref) + ". Would you like auto-calibration to "
            "take place automatically every time the keyboard is powered up?");

    if (result == QMessageBox::Yes)
    {
        diag.setVref(65535);
        diag.storeVref();
        diag.setVref(vref);
    }

    updateVref();
}

/*
 *
 */
void Frontend::storeVrefButtonClicked(void)
{
    diag.storeVref();
}

/*
 *
 */
void Frontend::buildMatrix(void)
{
    vector<vector<unsigned char>> scancodes = diag.scancodes();

    /* clear existing matrix widgets */
    QLayoutItem *item;
    while ((item = matrixGrid->takeAt(0)) != NULL)
    {
        delete item->widget();
        delete item;
    }
    keyWidgets.clear();

    int cols = scancodes.size();
    if (cols == 0)
        throw runtime_error("error: keyboard reports matrix with no columns");
    int rows = scancodes[0].size();

    keyWidgets = vector<vector<Key *>>(cols, vector<Key *>(rows));

    for (int col = 0; col < cols; col++)
    {
        QLabel *colLabel = new QLabel("<b>" + QString::number(col + 1) +
                "</b>");
        colLabel->setAlignment(Qt::AlignCenter);
        matrixGrid->addWidget(colLabel, 0, col + 1);

        for (int row = 0; row < rows; row++)
        {
            Key *key = new Key(diag, col, row, scancodes[col][row]);
            keyWidgets[col][row] = key;
            matrixGrid->addWidget(key, row + 1, col + 1);
        }
    }
}

/*
 *
 */
void Frontend::updateKeyStates(void)
{
    if (!isEnabled())
        return;

    vector<vector<bool>> states = diag.keyStates();

    int cols = states.size();
    if (cols == 0)
        throw runtime_error("error: keyboad reports key states with no "
                "columns");
    int rows = states[0].size();

    for (int col = 0; col < cols; col++)
        for (int row = 0; row < rows; row++)
            keyWidgets[col][row]->setPressed(states[col][row]);
}

/*
 *
 */
void Frontend::storeMatrixButtonClicked(void)
{
    setEnabled(false);
    storeMatrixButton->setText("Writing to EEPROM...");

    diag.storeScancodes();

    /* keyboard will disappear for a bit while it does its thing */
    QTimer::singleShot(1000, this, SLOT(storeMatrixComplete()));
}

/*
 *
 */
void Frontend::storeMatrixComplete(void)
{
    setEnabled(true);
    storeMatrixButton->setText("Store layout in EEPROM");
}

/*
 *
 */
void Frontend::loadMatrixButtonClicked(void)
{
    setEnabled(false);
    loadMatrixButton->setText("Reading from EEPROM...");

    diag.loadScancodes();

    /* keyboard will disappear for a (shorter) bit while it does its thing */
    QTimer::singleShot(250, this, SLOT(loadMatrixComplete()));
}

/*
 *
 */
void Frontend::loadMatrixComplete(void)
{
    setEnabled(true);
    loadMatrixButton->setText("Load layout from EEPROM");

    buildMatrix();
}

/*
 *
 */
void Frontend::importMatrixButtonClicked(void)
{
    QString filename = QFileDialog::getOpenFileName(this, "Import layout",
            QString(), "Layout files (*.l)");
    if (filename == QString())
        return;

    QString errStr;
    vector<vector<unsigned char>> scancodes;
    if (!importLayout(filename, scancodes, errStr))
    {
        QMessageBox::critical(this, "Error",
                QString("Could not import layout: ") + errStr);
        return;
    }

    try
    {
        diag.setScancodes(scancodes);
        buildMatrix();
    }
    catch (std::exception &e)
    {
        QMessageBox::critical(this, "Error", QString::fromUtf8(e.what()));
    }
}

/*
 *
 */
void Frontend::exportMatrixButtonClicked(void)
{
    QString filename = QFileDialog::getSaveFileName(this, "Export layout",
            "layout.l", "Layout files (*.l)");
    if (filename == QString())
        return;

    QString errStr;
    if (!exportLayout(filename, keyWidgets, errStr))
        QMessageBox::critical(this, "Error",
                QString("Could not export layout: ") + errStr);
}
